# Copyright (C) 2024  Johannes Sasongko <johannes sasongko org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

import dbm
import glob
import os
import os.path
import pickle
import shelve
from typing import List

try:
    import berkeleydb
except ImportError:
    import bsddb3 as berkeleydb

from xl import common, sqlitedbm


def migrate(old_path: str, new_path: str):
    """
    Migrate to SQLite-based shelf database.

    The source can be a BsdDbShelf but can also be any of the built-in shelves,
    because there was a time (probably only during 4.0 beta?) when Exaile could
    produce non-bsddb shelves.
    It even supports old databases created on Python 2.

    :return: Whether a migration is performed
    """

    class Utf8Unpickler(pickle.Unpickler):
        def __init__(self, *args, **kwargs):
            kwargs['encoding'] = 'utf-8'
            super().__init__(*args, **kwargs)

    # Change shelve's unpickler to use UTF-8, for compatibility with shelves
    # created by Python 2.
    # This is safe to remove if we don't want the compatibility anymore.
    shelve.Unpickler = Utf8Unpickler

    # Read data from old db
    try:
        old_db = berkeleydb.hashopen(old_path, 'r')
    except (berkeleydb.db.DBNoSuchFileError, berkeleydb.db.DBInvalidArgError):
        try:
            old_db = dbm.open(old_path, 'r')
        except dbm.error:
            return False
    old_shelf = shelve.Shelf(old_db, protocol=common.PICKLE_PROTOCOL)
    del old_db  # Don't accidentally use the raw database
    items = list(old_shelf.items())
    old_shelf.close()

    # Set aside the old files as temporary backup
    bak_path = old_path + '.before-sqlite'
    old_group = _get_file_group(old_path)
    bak_group = [bak_path + old[len(old_path) :] for old in old_group]
    for old, bak in zip(old_group, bak_group):
        os.rename(old, bak)

    # Write data to new db
    new_shelf = shelve.Shelf(
        sqlitedbm.SqliteDbm(new_path, autocommit=False), protocol=common.PICKLE_PROTOCOL
    )
    for key, value in items:
        new_shelf[key] = value
    new_shelf.close()

    # # Remove the backup files
    # for bak in _get_file_group(bak_path):
    #     try:
    #         os.remove(bak)
    #     except Exception:
    #         pass

    return True


def _get_file_group(path: str) -> List[str]:
    """
    Get `path` (if exists) and `path.*` files.

    Databases like to create multiple files; this tries to get all of them.
    """
    group = glob.glob(path + '.*')
    if os.path.exists(path):
        group.insert(0, path)
    return group



# winmmkeys - Adds support for multimedia keys on Windows
# Copyright (C) 2007, 2010, 2016, 2022  Johannes Sasongko <sasongko@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


class WinmmkeysPlugin:
    def enable(self, exaile):
        self.exaile = exaile

    def on_gui_loaded(self):
        self.handler = HotkeyHandler_Pynput(self.exaile)

    def disable(self, exaile):
        if hasattr(self, 'handler'):
            self.handler.disable()
            del self.handler
        del self.exaile


plugin_class = WinmmkeysPlugin


class HotkeyHandler_Pynput:
    def __init__(self, exaile):
        from pynput import keyboard

        from xl.player import PLAYER, QUEUE

        # https://learn.microsoft.com/windows/win32/inputdev/virtual-key-codes
        key_map = {
            0xB0: QUEUE.next,
            0xB1: QUEUE.prev,
            0xB2: PLAYER.stop,
            0xB3: PLAYER.toggle_pause,
            0xB5: exaile.gui.main.window.present,
        }

        # NOTE: This runs on every key event in the system, so it should exit as
        # fast as possible when the key or event type doesn't match.
        def on_event(msg, data):
            # WM_KEYDOWN = 0x0100
            if data.vkCode in key_map and msg == 0x0100:
                key_map[data.vkCode]()
                listener.suppress_event()

        self.listener = listener = keyboard.Listener(win32_event_filter=on_event)
        listener.start()

    def __del__(self):
        self.disable()

    def disable(self):
        if hasattr(self, 'listener'):
            self.listener.stop()
            del self.listener



from gi.repository import Gtk

from xl.common import idle_add, SimpleProgressThread
from xl.collection import Collection, Library, CollectionScanThread
from xl.nls import gettext as _
from xl.trax import search_tracks, TracksMatcher

from xlgui.progress import ProgressManager
from xlgui.widgets import dialogs

from xlgui.guiutil import GtkTemplate

from .gt_common import get_track_groups, set_track_groups

import logging

logger = logging.getLogger(__name__)


@GtkTemplate('gt_import.ui', relto=__file__)
class GtImporter(Gtk.Window):
    """
    Shows a dialog that allows importing of grouping tags
    from a directory not included in the current collection.
    """

    __gtype_name__ = 'GtImporter'

    (
        content_area,
        ok_button,
        tags_model,
        tags_view,
        tags_vbox,
        radio_merge,
        radio_replace,
    ) = GtkTemplate.Child.widgets(7)

    def __init__(self, exaile, uris):
        Gtk.Window.__init__(self, transient_for=exaile.gui.main.window)
        self.init_template()

        self.exaile = exaile

        self.collection = Collection("GT Import Collection")

        for uri in uris:
            self.collection.add_library(Library(uri))

        self.manager = ProgressManager(self.content_area)

        self.rescan_thread = CollectionScanThread(self.collection)
        self.rescan_thread.connect('done', self._on_rescan_done)

        self.import_thread = None

        self.manager.add_monitor(
            self.rescan_thread, _("Importing tracks"), 'document-open'
        )

    #
    # Status routines
    #

    @idle_add()
    def _on_rescan_done(self, thread):
        '''Called when the collection is finished loading'''

        if self.rescan_thread is None:
            return

        self.rescan_thread = None

        logger.info('Import directory scan completed, importing groups')

        # now that the collection has loaded, import the groups from them
        self.import_track_data = []
        self.import_thread = SimpleProgressThread(
            track_import_thread,
            self.collection,
            self.exaile.collection,
            self.import_track_data,
        )
        self.import_thread.connect('done', self._on_import_done)
        self.manager.add_monitor(
            self.import_thread, _("Importing groups"), 'document-open'
        )

    @idle_add()
    def _on_import_done(self, thread):
        '''Called when the grouping data is retrieved'''

        if self.import_thread is None:
            return

        track_data = self.import_track_data
        self.import_thread = None
        self.import_track_data = None

        logger.info('Group import finished, %s new tracks found', len(track_data))

        if len(track_data) == 0:
            self.destroy()

            locations = ';'.join(
                [l.get_location() for l in self.collection.get_libraries()]
            )

            dialogs.info(
                self.exaile.gui.main.window, 'No new tracks found at "%s"' % locations
            )
            return

        self.tags_view.freeze_child_notify()
        self.tags_view.set_model(None)

        # add the data to the model
        for old_group_str, new_group_str, matched_track, newgroups in track_data:
            self.tags_model.append(
                (
                    True,
                    str(matched_track),
                    old_group_str,
                    new_group_str,
                    matched_track,
                    newgroups,
                )
            )

        self.tags_view.set_model(self.tags_model)
        self.tags_view.thaw_child_notify()

        self.ok_button.set_sensitive(True)
        self.tags_vbox.set_visible(True)

    @idle_add()
    def _on_update_done(self, thread):
        if self.update_thread is None:
            return

        self.update_thread = None

        logger.info('Track update complete')
        self.destroy()

    #
    # Widget events
    #

    @GtkTemplate.Callback
    def on_cancel_button_clicked(self, widget):
        if self.rescan_thread is not None:
            self.rescan_thread.stop()
        elif self.import_thread is not None:
            self.import_thread.stop()
        elif self.update_thread is not None:
            self.update_thread.stop()

        self.destroy()

    @GtkTemplate.Callback
    def on_import_checkbox_toggled(self, cell, path):
        self.tags_model[path][0] = not cell.get_active()

    @GtkTemplate.Callback
    def on_ok_button_clicked(self, widget):
        self.ok_button.set_sensitive(False)
        self.tags_vbox.set_sensitive(False)

        data = [(row[4], row[5]) for row in self.tags_model if row[0]]
        logger.info('Updating %s tracks', len(data))

        self.update_thread = SimpleProgressThread(
            track_update_thread, data, self.radio_replace.get_active()
        )
        self.update_thread.connect('done', self._on_update_done)

        self.manager.add_monitor(self.update_thread, _("Updating groups"), 'system-run')

    @GtkTemplate.Callback
    def on_window_destroy(self, widget):
        self.collection.close()


def track_import_thread(import_collection, user_collection, track_data):
    """
    Reads grouping information from tracks in a collection, and matches
    them with tracks contained in a separate collection.
    """

    total = float(len(import_collection))

    # to import, all essential fields should be identical!
    fields = ['__length', 'artist', 'album', 'title', 'genre', 'tracknumber']

    logger.info("Finding matches for %s imported tracks", len(import_collection))

    exact_dups = 0
    no_change = 0

    # determine which tracks in this collection match existing
    # tracks in the exaile collection. grab the groups from them
    for i, track in enumerate(import_collection):
        # search for a matching track
        # -> currently exaile doesn't index tracks, and linear searches
        #    for the track instead. Oh well.

        matchers = [TracksMatcher(track.get_tag_search(t)) for t in fields]
        matched_tracks = [r.track for r in search_tracks(user_collection, matchers)]

        # if there are matches, add the data to the track data
        for matched_track in matched_tracks:
            # ignore exact duplicates
            if track is matched_track:
                exact_dups += 1
                continue

            old_group_str = ' '.join(get_track_groups(matched_track))
            newgroups = get_track_groups(track)
            new_group_str = ' '.join(newgroups)

            if old_group_str == new_group_str:
                no_change += 1
                continue

            track_data.append((old_group_str, new_group_str, matched_track, newgroups))

        yield (i, total)

    logger.info(
        "Match information: %s exact dups, %s no change, %s differing tracks",
        exact_dups,
        no_change,
        len(track_data),
    )


def track_update_thread(trackdata, replace):
    """
    Sets new groups on a set of tracks
    """
    total = len(trackdata)

    for i, (curtrack, newgroups) in enumerate(trackdata):
        if replace:
            set_track_groups(curtrack, newgroups)
        else:
            curgroups = get_track_groups(curtrack) | newgroups
            set_track_groups(curtrack, curgroups)

        yield (i, total)


def import_tags(exaile):
    """
    Function to show a dialog that allows the user to import grouping
    tags from a directory of their choosing.
    """

    def _on_uris_selected(widget, uris):
        import_dialog = GtImporter(exaile, uris)
        import_dialog.show()

    file_dialog = dialogs.DirectoryOpenDialog(
        exaile.gui.main.window, title=_('Select directory to import grouping tags from')
    )
    file_dialog.connect('uris-selected', _on_uris_selected)
    file_dialog.run()
    file_dialog.destroy()



# Python console plugin for Exaile media player
# Copyright (C) 2007 Johannes Sasongko
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""Simple console plugin

For better development experience, please use the IPython console plugin. This
plugin is meant as a basic alternative without the extra dependencies.
"""

from gi.repository import Gtk

from contextlib import redirect_stdout
from io import StringIO
import os
import sys
import traceback


class PyConsole:
    def __init__(self, dict, exaile):
        self.dict = dict
        self.buffer = StringIO()

        ui = Gtk.Builder()
        ui.add_from_file(
            os.path.join(
                os.path.dirname(os.path.realpath(__file__)), 'console_window.ui'
            )
        )

        self.window = ui.get_object('simple_console_window')
        self.close_handler = self.window.connect(
            'delete-event', console_destroyed, exaile
        )

        self.text_view = tv = ui.get_object('console_output')

        self.text_buffer = buff = tv.get_buffer()
        self.end_mark = buff.create_mark(None, buff.get_end_iter(), False)

        self.entry = entry = ui.get_object('console_input')
        entry.connect('activate', self.entry_activated)
        entry.grab_focus()

    def entry_activated(self, entry, user_data=None):
        """
        Called when the user presses Return on the GtkEntry.
        """
        self.execute(entry.get_text())
        entry.select_region(0, -1)

    def execute(self, code):
        """
        Executes some Python code.
        """
        buffer = self.buffer
        try:
            pycode = compile(code, '<console>', 'single')
            with redirect_stdout(buffer):
                exec(pycode, self.dict)
        except Exception:
            exc = traceback.format_exception(*sys.exc_info())
            del exc[1]  # Remove our function.
            result = ''.join(exc)
        else:
            result = buffer.getvalue()
            # Can't simply close and recreate later because help() stores and
            # reuses stdout.
            buffer.seek(0)
            buffer.truncate(0)
        result = '>>> %s\n%s' % (code, result)
        self.text_buffer.insert(self.text_buffer.get_end_iter(), result)
        # Can't use iter; won't scroll correctly.
        self.text_view.scroll_to_mark(self.end_mark, 0, False, 0.5, 0.5)
        self.entry.grab_focus()


PLUGIN = None


def enable(exaile):
    if exaile.loading:
        from xl import event

        event.add_callback(_enable, 'gui_loaded')
    else:
        _enable(None, exaile, None)


def _enable(eventname, exaile, eventdata):
    global PLUGIN
    PLUGIN = PyConsole({'exaile': exaile}, exaile)
    PLUGIN.window.set_transient_for(exaile.gui.main.window)
    PLUGIN.window.present()


def console_destroyed(window, event, exaile):
    """Disable plugin on window destroy"""
    global PLUGIN
    if PLUGIN:
        exaile.plugins.disable_plugin(__name__)


def disable(exaile):
    global PLUGIN
    if PLUGIN:
        PLUGIN.window.disconnect(PLUGIN.close_handler)
        PLUGIN.window.destroy()
        PLUGIN = None


# vi: et sts=4 sw=4 ts=4



"""
This module is a parser for audio CDs based on discid.
It is compatible with both python-discid and python-libdiscid.
"""

try:  # allow both python-discid and python-libdiscid
    from libdiscid.compat import discid
except ImportError:
    import discid

from xl.trax import Track


def read_disc_id(device):
    """
    I/O operation to read an ID from the disc.
    This must happen async because it may take quite some time.

    @param device: Name of the CD device
    @return: The disc ID as understood by musicbrainz and parse_disc()
    """
    try:  # retry with reduced features if it fails
        # Note: reading additional features like isrc will take some time.
        disc_id = discid.read(device, features=discid.FEATURES)
    except discid.DiscError:
        disc_id = discid.read(device)
    return disc_id


def parse_disc(disc_id, device):
    """
    Parses the given disc ID into tracks.
    As a result, the data will only contain track numbers and lengths but
    no sophisticated metadata.

    @param disc_id: The disc ID from read_disc_id()
    @param device: Name of the CD device
    @return: An array of xl.trax.Track with minimal information
    """
    xl_tracks = []
    disc_tags = dict()
    # The tag name is chosen for compatibility with the cover manager!
    disc_tags['musicbrainz_albumid'] = disc_id.id
    disc_tags['__freedb_disc_id'] = disc_id.freedb_id
    if disc_id.mcn and '0000000000000' not in disc_id.mcn:
        disc_tags['__mcn'] = disc_id.mcn

    for discid_track in disc_id.tracks:
        track_tags = disc_tags.copy()
        track_tags['tracknumber'] = '{0}/{1}'.format(
            discid_track.number, len(disc_id.tracks)
        )
        track_tags['title'] = "Track %d" % discid_track.number
        track_tags['__length'] = discid_track.seconds
        if discid_track.isrc:
            track_tags['isrc'] = discid_track.isrc

        track_uri = "cdda://%d/#%s" % (discid_track.number, device)
        track = Track(uri=track_uri, scan=False)
        track.set_tags(**track_tags)
        xl_tracks.append(track)
    return xl_tracks



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

from gi.repository import Atk
from gi.repository import Gdk
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import Pango

from xl import event, formatter, player, providers, settings
from xl.common import clamp
from xl.nls import gettext as _
from xlgui.widgets import menu

from xlgui.guiutil import GtkTemplate

import logging

logger = logging.getLogger(__name__)


class ProgressBarFormatter(formatter.ProgressTextFormatter):
    """
    A formatter for progress bars
    """

    def __init__(self, player):
        formatter.ProgressTextFormatter.__init__(self, '', player)

        self.on_option_set('gui_option_set', settings, 'gui/progress_bar_text_format')
        event.add_ui_callback(self.on_option_set, 'gui_option_set')

    def on_option_set(self, event, settings, option):
        """
        Updates the internal format on setting change
        """
        if option == 'gui/progress_bar_text_format':
            self.props.format = settings.get_option(
                'gui/progress_bar_text_format', '$current_time / $remaining_time'
            )


class PlaybackProgressBar(Gtk.ProgressBar):
    """
    Progress bar which automatically follows playback
    """

    def __init__(self, player):
        Gtk.ProgressBar.__init__(self)
        self.__player = player

        try:
            awidget = self.get_accessible()
            awidget.set_role(Atk.Role.AUDIO)
        except Exception:
            logger.debug("Exception setting ATK role", exc_info=True)

        self.set_show_text(True)

        self.reset()

        self.formatter = ProgressBarFormatter(player)
        self.__timer_id = None
        self.__events = (
            'playback_track_start',
            'playback_player_end',
            'playback_toggle_pause',
            'playback_error',
        )

        for e in self.__events:
            event.add_ui_callback(getattr(self, 'on_%s' % e), e, self.__player)

    def destroy(self):
        """
        Cleanups
        """
        for e in self.__events:
            event.remove_callback(getattr(self, 'on_%s' % e), e, self.__player)

    def reset(self):
        """
        Resets the progress bar appearance
        """
        self.set_fraction(0)
        self.set_text(_('Not Playing'))

    def __enable_timer(self):
        """
        Enables the update timer
        """
        if self.__timer_id is not None:
            return

        interval = settings.get_option('gui/progress_update_millisecs', 1000)

        if interval % 1000 == 0:
            self.__timer_id = GLib.timeout_add_seconds(interval // 1000, self.on_timer)
        else:
            self.__timer_id = GLib.timeout_add(interval, self.on_timer)

        self.on_timer()

    def __disable_timer(self):
        """
        Disables the update timer
        """
        if self.__timer_id is not None:
            GLib.source_remove(self.__timer_id)
            self.__timer_id = None

    def on_timer(self):
        """
        Updates progress bar appearance
        """
        if self.__player.current is None:
            self.__disable_timer()
            self.reset()
            return False

        self.set_fraction(self.__player.get_progress())
        self.set_text(self.formatter.format())

        return True

    def on_playback_track_start(self, event_type, player, track):
        """
        Starts update timer
        """
        self.reset()
        self.__enable_timer()

    def on_playback_player_end(self, event_type, player, track):
        """
        Stops update timer
        """
        self.__disable_timer()
        self.reset()

    def on_playback_toggle_pause(self, event_type, player, track):
        """
        Starts or stops update timer
        """
        if player.is_playing():
            self.__enable_timer()
        elif player.is_paused():
            self.__disable_timer()

    def on_playback_error(self, event_type, player, message):
        """
        Stops update timer
        """
        self.__disable_timer()
        self.reset()


class Anchor(int):
    __gtype__ = GObject.TYPE_INT


for i, a in enumerate(
    'CENTER NORTH NORTH_WEST NORTH_EAST SOUTH SOUTH_WEST SOUTH_EAST WEST EAST'.split()
):
    setattr(Anchor, a, Anchor(i))


class Marker(GObject.Object):
    """
    A marker pointing to a playback position
    """

    __gproperties__ = {
        'anchor': (
            Anchor,
            'anchor position',
            'The position the marker will be anchored',
            Anchor.CENTER,
            Anchor.EAST,
            Anchor.SOUTH,
            GObject.ParamFlags.READWRITE,
        ),
        'color': (
            Gdk.RGBA,
            'marker color',
            'Override color of the marker',
            GObject.ParamFlags.READWRITE,
        ),
        'label': (
            GObject.TYPE_STRING,
            'marker label',
            'Textual description of the marker',
            None,
            GObject.ParamFlags.READWRITE,
        ),
        'position': (
            GObject.TYPE_FLOAT,
            'marker position',
            'Relative position of the marker',
            0,
            1,
            0,
            GObject.ParamFlags.READWRITE,
        ),
        'state': (
            Gtk.StateType,
            'marker state',
            'The state of the marker',
            Gtk.StateType.NORMAL,
            GObject.ParamFlags.READWRITE,
        ),
    }
    __gsignals__ = {'reached': (GObject.SignalFlags.RUN_LAST, None, ())}

    def __init__(self, position: float = 0):
        GObject.Object.__init__(self)

        self.__values = {
            'anchor': Anchor.SOUTH,
            'color': None,
            'label': None,
            'position': 0,
            'state': Gtk.StateType.NORMAL,
        }

        self.props.position = position

    def __str__(self):
        """
        Informal representation
        """
        if self.props.label is not None:
            text = '%s (%g)' % (self.props.label, self.props.position)
        else:
            text = '%g' % self.props.position

        return text

    def __lt__(self, other):
        """
        Compares positions
        """
        return self.props.position < other.props.position

    def __gt__(self, other):
        """
        Compares positions
        """
        return self.props.position > other.props.position

    def do_get_property(self, gproperty):
        """
        Gets a GObject property
        """
        try:
            return self.__values[gproperty.name]
        except KeyError:
            raise AttributeError('unknown property %s' % gproperty.name)

    def do_set_property(self, gproperty, value):
        """
        Sets a GObject property
        """
        if gproperty.name not in self.__values:
            raise AttributeError('unknown property %s' % gproperty.name)
        self.__values[gproperty.name] = value


class MarkerManager(providers.ProviderHandler):
    """
    Enables management of playback markers; namely simple
    adding, removing and finding. It also takes care of
    emitting signals when a marker is reached during playback.

    TODO: This presumes there is only one player object present
    in exaile, and that markers can only be associated with
    the single player object. This class should probably be
    changed to be associated with a particular player (which
    requires some changes to the marker class)
    """

    def __init__(self):
        providers.ProviderHandler.__init__(self, 'playback-markers')

        self.__events = ('playback_track_start', 'playback_track_end')
        self.__timeout_id = None

        for e in self.__events:
            event.add_ui_callback(getattr(self, 'on_%s' % e), e)

    def destroy(self):
        """
        Cleanups
        """
        for e in self.__events:
            event.remove_callback(getattr(self, 'on_%s' % e), e)

    def add_marker(self, position: float) -> Marker:
        """
        Creates a new marker for a playback position

        :param position: the playback position [0..1]
        :returns: the new marker
        """
        marker = Marker(position)
        # Provider compatibility
        marker.name = 'marker'
        providers.register('playback-markers', marker)

        return marker

    def remove_marker(self, marker: Marker) -> None:
        """
        Removes a playback marker

        :param marker: the marker
        """
        providers.unregister('playback-markers', marker)

    def get_markers_at(self, position):
        """
        Gets all markers located at a position

        :param position: the mark position
        :type position: float
        :returns: (m1, m2, ...)
        :rtype: (:class:`Marker`, ...)

        * *m1*: the first marker
        * *m2*: the second marker
        * ...
        """
        # Reproduce value modifications
        position = Marker(position).props.position
        markers = ()

        for marker in providers.get('playback-markers'):
            if marker.props.position == position:
                markers += (marker,)

        return markers

    def on_playback_track_start(self, event, player, track):
        """
        Starts marker watching
        """
        if self.__timeout_id is not None:
            GLib.source_remove(self.__timeout_id)

        self.__timeout_id = GLib.timeout_add_seconds(1, self.on_timeout, player)

    def on_playback_track_end(self, event, player, track):
        """
        Stops marker watching
        """
        if self.__timeout_id is not None:
            GLib.source_remove(self.__timeout_id)
            self.__timeout_id = None

    def on_timeout(self, player):
        """
        Triggers "reached" signal of markers
        """

        if player.current is None:
            self.__timeout_id = None
            return

        track_length = player.current.get_tag_raw('__length')

        if track_length is None:
            return True

        playback_time = int(player.get_time())
        reached_markers = (
            m
            for m in providers.get('playback-markers')
            if int(m.props.position * track_length) == playback_time
        )

        for marker in reached_markers:
            marker.emit('reached')

        return True


__MARKERMANAGER = MarkerManager()
add_marker = __MARKERMANAGER.add_marker
remove_marker = __MARKERMANAGER.remove_marker
get_markers_at = __MARKERMANAGER.get_markers_at


class _SeekInternalProgressBar(PlaybackProgressBar):
    def __init__(self, player, points, marker_scale):
        PlaybackProgressBar.__init__(self, player)
        self._points = points
        self._seeking = False
        self._marker_scale = marker_scale

    def do_draw(self, context):
        """
        Draws markers on top of the progress bar
        """
        Gtk.ProgressBar.do_draw(self, context)

        if not self._points:
            return

        context.set_line_width(self._marker_scale / 0.9)
        style = self.get_style_context()

        POW_256_2 = 256**2

        for marker, points in self._points.items():
            for i, (x, y) in enumerate(points):
                if i == 0:
                    context.move_to(x, y)
                else:
                    context.line_to(x, y)
            context.close_path()

            if marker.props.state in (Gtk.StateType.PRELIGHT, Gtk.StateType.ACTIVE):
                c = style.get_color(Gtk.StateFlags.NORMAL)
                context.set_source_rgba(c.red, c.green, c.blue, c.alpha)
            else:
                if marker.props.color is not None:
                    base = marker.props.color
                else:
                    base = style.get_color(marker.props.state)

                context.set_source_rgba(
                    base.red / POW_256_2,
                    base.green / POW_256_2,
                    base.blue / POW_256_2,
                    0.7,
                )
            context.fill_preserve()

            if marker.props.state in (Gtk.StateType.PRELIGHT, Gtk.StateType.ACTIVE):
                c = style.get_color(Gtk.StateFlags.NORMAL)
                context.set_source_rgba(c.red, c.green, c.blue, c.alpha)
            else:
                foreground = style.get_color(marker.props.state)
                context.set_source_rgba(
                    foreground.red / POW_256_2,
                    foreground.green / POW_256_2,
                    foreground.blue / POW_256_2,
                    0.7,
                )
            context.stroke()

    def on_timer(self):
        """
        Prevents update while seeking
        """
        if self._seeking:
            return True

        return PlaybackProgressBar.on_timer(self)


class SeekProgressBar(Gtk.EventBox, providers.ProviderHandler):
    """
    Playback progress bar which allows for seeking
    and setting positional markers
    """

    CSS = Gtk.CssProvider()
    CSS.load_from_data(b'''
        /* Make the text easier to read on Adwaita */
        progressbar {
            color: unset;
            font-size: unset;
        }
        ''')

    __gproperties__ = {
        'marker-scale': (
            GObject.TYPE_FLOAT,
            'marker scale',
            'Scaling of markers',
            0,
            1,
            0.7,
            GObject.ParamFlags.READWRITE,
        )
    }
    __gsignals__ = {
        'button-press-event': 'override',
        'button-release-event': 'override',
        'motion-notify-event': 'override',
        'notify': 'override',
        'key-press-event': 'override',
        'key-release-event': 'override',
        'scroll-event': 'override',
        'marker-reached': (
            GObject.SignalFlags.RUN_LAST,
            GObject.TYPE_BOOLEAN,
            (Marker,),
            GObject.signal_accumulator_true_handled,
        ),
    }

    def __init__(self, player, use_markers=True):
        """
        TODO: markers aren't designed for more than one player, once
        they are we can get rid of the use_markers option
        """

        Gtk.EventBox.__init__(self)

        points = {}

        self.__player = player
        self.__values = {'marker-scale': 0.7}
        self._points = points

        self.__progressbar = _SeekInternalProgressBar(
            player, points, self.__values['marker-scale']
        )
        self._progressbar_menu = None

        if use_markers:
            self._progressbar_menu = ProgressBarContextMenu(self)

            self._marker_menu = MarkerContextMenu(self)
            self._marker_menu.connect('deactivate', self.on_marker_menu_deactivate)

            providers.ProviderHandler.__init__(self, 'playback-markers')

        sc = self.__progressbar.get_style_context()
        sc.add_provider(self.CSS, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)

        self.add_events(
            Gdk.EventMask.BUTTON_PRESS_MASK
            | Gdk.EventMask.BUTTON_RELEASE_MASK
            | Gdk.EventMask.POINTER_MOTION_MASK
            | Gdk.EventMask.LEAVE_NOTIFY_MASK
            | Gdk.EventMask.SCROLL_MASK
        )

        self.set_can_focus(True)

        self.connect('hierarchy-changed', self.on_hierarchy_changed)
        self.connect('scroll-event', self.on_scroll_event)

        self.add(self.__progressbar)
        self.show_all()

    def get_label(self, marker):
        """
        Builds the most appropriate label
        markup to describe a marker

        :param marker: the marker
        :type marker: :class:`Marker`
        :returns: the label
        :rtype: string
        """
        markup = None

        if self.__player.current:
            length = self.__player.current.get_tag_raw('__length')

            if length is not None:
                length = length * marker.props.position
                length = formatter.LengthTagFormatter.format_value(length)

                if marker.props.label:
                    markup = '<b>%s</b> (%s)' % (marker.props.label, length)
                else:
                    markup = '%s' % length
        else:
            if marker.props.label:
                markup = '<b>%s</b> (%d%%)' % (
                    marker.props.label,
                    int(marker.props.position * 100),
                )
            else:
                markup = '%d%%' % int(marker.props.position * 100)

        return markup

    def _is_marker_hit(self, marker, check_x, check_y):
        """
        Checks whether a marker is hit by a point

        :param marker: the marker
        :type marker: :class:`Marker`
        :param check_x: the x location to check
        :type check_x: float
        :param check_y: the y location to check
        :type check_y: float
        :returns: whether the marker was hit
        :rtype: bool
        """
        points = self._points[marker]
        x, y, width, height = self._get_bounding_box(points)

        if x <= check_x <= width and y <= check_y <= height:
            return True

        return False

    def _get_points(self, marker, width=None, height=None):
        """
        Calculates the points necessary
        to represent a marker

        :param marker: the marker
        :type marker: :class:`Marker`
        :param width: area width override
        :type width: int
        :param height: area height override
        :type height: int
        :returns: ((x1, y1), (x2, y2), ...)
        :rtype: ((float, float), ...)

        * *x1*: the x coordinate of the first point
        * *y1*: the y coordinate of the first point
        * *x2*: the x coordinate of the second point
        * *y2*: the y coordinate of the second point
        * ...
        """
        points = ()
        alloc = self.get_allocation()
        width = width or alloc.width
        height = height or alloc.height
        position = width * marker.props.position
        marker_scale = int(height * self.props.marker_scale)
        # Adjustment by half of the line width
        offset = self.props.marker_scale / 0.9 / 2

        if marker.props.anchor == Anchor.NORTH_WEST:
            points = (
                (position - offset, offset),
                (position + marker_scale * 0.75 - offset, offset),
                (position - offset, marker_scale * 0.75 + offset),
            )
        elif marker.props.anchor == Anchor.NORTH:
            points = (
                (position - offset, marker_scale // 2 + offset),
                (position + marker_scale // 2 - offset, offset),
                (position - marker_scale // 2 - offset, offset),
            )
        elif marker.props.anchor == Anchor.NORTH_EAST:
            points = (
                (position - marker_scale * 0.75 - offset, offset),
                (position - offset, offset),
                (position - offset, marker_scale * 0.75 + offset),
            )
        elif marker.props.anchor == Anchor.EAST:
            points = (
                (position - marker_scale // 2 - offset, height // 2 + offset),
                (position - offset, height // 2 - marker_scale // 2 + offset),
                (position - offset, height // 2 + marker_scale // 2 + offset),
            )
        elif marker.props.anchor == Anchor.SOUTH_EAST:
            points = (
                (position - offset, height - offset),
                (position - offset, height - marker_scale * 0.75 - offset),
                (position - marker_scale * 0.75 - offset, height - offset),
            )
        elif marker.props.anchor == Anchor.SOUTH:
            points = (
                (position - offset, height - marker_scale // 2 - offset),
                (position + marker_scale // 2 - offset, height - offset),
                (position - marker_scale // 2 - offset, height - offset),
            )
        elif marker.props.anchor == Anchor.SOUTH_WEST:
            points = (
                (position - offset, height - offset),
                (position + marker_scale * 0.75 - offset, height - offset),
                (position - offset, height - marker_scale * 0.75 - offset),
            )
        elif marker.props.anchor == Anchor.WEST:
            points = (
                (position + marker_scale // 2 - offset, height // 2 + offset),
                (position - offset, height // 2 - marker_scale // 2 + offset),
                (position - offset, height // 2 + marker_scale // 2 + offset),
            )
        elif marker.props.anchor == Anchor.CENTER:
            points = (
                (position - offset, height // 2 - marker_scale // 2 + offset),
                (position + marker_scale // 2 - offset, height // 2 + offset),
                (position - offset, height // 2 + marker_scale // 2 + offset),
                (position - marker_scale // 2 - offset, height // 2 + offset),
            )

        return points

    def _get_bounding_box(self, points):
        """
        Calculates the axis aligned bounding box
        of a sequence of points

        :param points: ((x1, y1), (x2, y2), ...)
        :type points: ((float, float), ...)
        :returns: (x, y, width, height)
        :rtype: (float, float, float, float)

        * *x*: the x coordinate of the box
        * *y*: the y coordinate of the box
        * *width*: the width of the box
        * *height*: the height of the box
        """
        xs, ys = zip(*points)
        return min(xs), min(ys), max(xs), max(ys)

    def seek(self, position):
        """
        Seeks within the current track
        """
        if self.__player.current:
            self.__player.set_progress(position)
            self.update_progress()

    def update_progress(self):
        """
        Updates the progress bar and the time with data from the player
        """

        if self.__player.current:
            length = self.__player.current.get_tag_raw('__length')

            if length is not None:
                position = float(self.__player.get_time()) / length
                self.__progressbar.set_fraction(position)
                self.__progressbar.set_text(
                    self.__progressbar.formatter.format(current_time=length * position)
                )

    def do_get_property(self, gproperty):
        """
        Gets a GObject property
        """
        try:
            return self.__values[gproperty.name]
        except KeyError:
            raise AttributeError('unknown property %s' % property.name)

    def do_set_property(self, gproperty, value):
        """
        Sets a GObject property
        """
        try:
            self.__values[gproperty.name] = value
        except KeyError:
            raise AttributeError('unknown property %s' % property.name)

    def do_notify(self, gproperty):
        """
        Reacts to GObject property changes
        """
        if gproperty.name == 'marker-scale':
            for marker in self._points:
                self._points[marker] = self._get_points(marker)
            self.__progressbar._marker_scale = self.__values['marker-scale']
            self.__progressbar.queue_draw()

    def do_size_allocate(self, allocation):
        """
        Recalculates the marker points
        """
        oldallocation = self.get_allocation()

        Gtk.EventBox.do_size_allocate(self, allocation)

        if allocation != oldallocation:
            for marker in self._points:
                self._points[marker] = self._get_points(marker)

    def do_button_press_event(self, event):
        """
        Prepares seeking
        """
        event = event.button
        hit_markers = []

        for marker in self._points:
            if self._is_marker_hit(marker, event.x, event.y):
                if marker.props.state in (Gtk.StateType.NORMAL, Gtk.StateType.PRELIGHT):
                    marker.props.state = Gtk.StateType.ACTIVE
                    hit_markers += [marker]

        hit_markers.sort()

        if event.button == Gdk.BUTTON_PRIMARY:
            if self.__player.current is None:
                return True

            length = self.__player.current.get_tag_raw('__length')

            if length is None:
                return True

            if len(hit_markers) > 0:
                self.seek(hit_markers[0].props.position)
            else:
                fraction = event.x / self.get_allocation().width
                fraction = clamp(fraction, 0, 1)

                self.__progressbar.set_fraction(fraction)
                self.__progressbar.set_text(
                    _('Seeking: %s')
                    % self.__progressbar.formatter.format(
                        current_time=length * fraction
                    )
                )
                self.__progressbar._seeking = True
        elif event.triggers_context_menu():
            if len(hit_markers) > 0:
                self._marker_menu.popup(event, tuple(hit_markers))
            elif self._progressbar_menu is not None:
                self._progressbar_menu.popup(event)

    def do_button_release_event(self, event):
        """
        Completes seeking
        """
        event = event.button

        for marker in self._points:
            if marker.props.state == Gtk.StateType.ACTIVE:
                marker.props.state = Gtk.StateType.PRELIGHT

        if event.button == Gdk.BUTTON_PRIMARY and self.__progressbar._seeking:
            fraction = event.x / self.get_allocation().width
            fraction = clamp(fraction, 0, 1)

            self.seek(fraction)
            self.__progressbar._seeking = False

    def do_motion_notify_event(self, event):
        """
        Updates progress bar while seeking
        and updates marker states on hover
        """
        self.set_tooltip_markup(None)

        if self.__progressbar._seeking:
            press_event = Gdk.EventButton.new(Gdk.EventType.BUTTON_PRESS)
            press_event.button = Gdk.BUTTON_PRIMARY
            press_event.x = event.x
            press_event.y = event.y

            self.emit('button-press-event', press_event)
        else:
            hit_markers = []

            for marker in self._points:
                if self._is_marker_hit(marker, event.x, event.y):
                    if marker.props.state == Gtk.StateType.NORMAL:
                        marker.props.state = Gtk.StateType.PRELIGHT
                    hit_markers += [marker]
                else:
                    if marker.props.state == Gtk.StateType.PRELIGHT:
                        marker.props.state = Gtk.StateType.NORMAL

            if len(hit_markers) > 0:
                hit_markers.sort()
                markup = ', '.join([self.get_label(m) for m in hit_markers])
                self.set_tooltip_markup(markup)
                self.trigger_tooltip_query()

    def do_leave_notify_event(self, event):
        """
        Resets marker states
        """
        for marker in self._points:
            # Leave other states intact
            if marker.props.state == Gtk.StateType.PRELIGHT:
                marker.props.state = Gtk.StateType.NORMAL

    def do_key_press_event(self, event):
        """
        Prepares seeking via keyboard interaction
        * Alt+Up/Right: seek 1% forward
        * Alt+Down/Left: seek 1% backward
        """
        _, state = event.get_state()
        if state & Gtk.StateType.INSENSITIVE:
            return
        if not state & Gdk.ModifierType.MOD1_MASK:
            return

        if event.keyval in (Gdk.KEY_Up, Gdk.KEY_Right):
            direction = 1
        elif event.keyval in (Gdk.KEY_Down, Gdk.KEY_Left):
            direction = -1
        else:
            return

        press_event = Gdk.Event.new(Gdk.EventType.BUTTON_PRESS)
        press_event.button = Gdk.BUTTON_PRIMARY
        new_fraction = self.__progressbar.get_fraction() + 0.01 * direction
        alloc = self.get_allocation()
        press_event.x = alloc.width * new_fraction
        press_event.y = float(alloc.y)

        self.emit('button-press-event', press_event)

    def do_key_release_event(self, event):
        """
        Completes seeking via keyboard interaction
        """
        _, state = event.get_state()
        if not state & Gdk.ModifierType.MOD1_MASK:
            return

        if event.keyval in (Gdk.KEY_Up, Gdk.KEY_Right):
            direction = 1
        elif event.keyval in (Gdk.KEY_Down, Gdk.KEY_Left):
            direction = -1
        else:
            return

        release_event = Gdk.Event.new(Gdk.EventType.BUTTON_RELEASE)
        release_event.button = Gdk.BUTTON_PRIMARY
        new_fraction = self.__progressbar.get_fraction() + 0.01 * direction
        alloc = self.get_allocation()
        release_event.x = alloc.width * new_fraction
        release_event.y = float(alloc.y)

        self.emit('button-release-event', release_event)

    def on_scroll_event(self, widget, event):
        """
        Seek on scroll as VLC does
        """
        if not self.__player.current:
            return True
        if self.__player.current.get_tag_raw('__length') is None:
            return True

        progress = self.__player.get_progress()
        progress_delta = 0.05  # 5% of track length
        progress_delta_small = 0.005  # 0.5% of track length

        if (
            event.direction == Gdk.ScrollDirection.DOWN
            or event.direction == Gdk.ScrollDirection.LEFT
        ):
            if event.get_state() & Gdk.ModifierType.SHIFT_MASK:
                new_progress = progress - progress_delta_small
            else:
                new_progress = progress - progress_delta
        elif (
            event.direction == Gdk.ScrollDirection.UP
            or event.direction == Gdk.ScrollDirection.RIGHT
        ):
            if event.get_state() & Gdk.ModifierType.SHIFT_MASK:
                new_progress = progress + progress_delta_small
            else:
                new_progress = progress + progress_delta
        elif event.direction == Gdk.ScrollDirection.SMOOTH:
            if event.get_state() & Gdk.ModifierType.SHIFT_MASK:
                new_progress = progress + progress_delta_small * (
                    event.deltax + event.deltay
                )
            else:
                new_progress = progress + progress_delta * (event.deltax + event.deltay)

        self.__player.set_progress(clamp(new_progress, 0, 1))
        self.update_progress()

        return True

    def on_hierarchy_changed(self, widget, old_toplevel):
        """
        Sets up editing cancel on toplevel focus out
        """
        # Disconnect from previous toplevel.
        prev_conn = getattr(self, '_SeekProgressBar__prev_focus_out_conn', None)
        if prev_conn:
            prev_conn[0].disconnect(prev_conn[1])
            del self.__prev_focus_out_conn

        # Connect to new toplevel and store the connection, but only if it's an
        # actual toplevel window.
        toplevel = self.get_toplevel()
        if toplevel.is_toplevel():
            conn = toplevel.connect(
                'focus-out-event', lambda w, e: self.emit('focus-out-event', e.copy())
            )
            self.__prev_focus_out_conn = (toplevel, conn)

    def on_marker_menu_deactivate(self, menu):
        """
        Makes sure to reset states of
        previously selected markers
        """
        for marker in self._points:
            marker.props.state = Gtk.StateType.NORMAL
        self.__progressbar.queue_draw()

    def on_marker_notify(self, marker, gproperty):
        """
        Recalculates marker points on position changes
        """
        if gproperty.name in ('anchor', 'position'):
            self._points[marker] = self._get_points(marker)
        self.__progressbar.queue_draw()

    def on_provider_added(self, marker):
        """
        Calculates points after marker addition

        :param marker: the new marker
        :type marker: :class:`Marker`
        """
        notify_id = marker.connect('notify', self.on_marker_notify)
        setattr(marker, '%s_notify_id' % id(self), notify_id)
        self._points[marker] = self._get_points(marker)
        self.__progressbar.queue_draw()

    def on_provider_removed(self, marker):
        """
        Removes points from internal cache

        :param marker: the marker
        :type marker: :class:`Marker`
        """
        notify_id = getattr(marker, '%s_notify_id' % id(self))
        if notify_id is not None:
            marker.disconnect(notify_id)

        del self._points[marker]
        self.__progressbar.queue_draw()

    # HACK: These methods implement the PlaybackAdapter interface (passing the
    # calls to the internal progress bar, which is an actual PlaybackAdapter).
    # This class only pretends to be a PlaybackAdapter because we don't want
    # the mixin behavior here.

    def on_playback_track_start(self, event, player, track):
        self.__progressbar.on_playback_track_start(event, player, track)

    def on_playback_track_end(self, event, player, track):
        self.__progressbar.on_playback_track_end(event, player, track)

    def on_playback_player_end(self, event, player, track):
        self.__progressbar.on_playback_player_end(event, player, track)

    def on_playback_toggle_pause(self, event, player, track):
        self.__progressbar.on_playback_toggle_pause(event, player, track)

    def on_playback_error(self, event, player, message):
        self.__progressbar.on_playback_error(event, player, message)


class ProgressBarContextMenu(menu.ProviderMenu):
    """
    Progress bar specific context menu
    """

    def __init__(self, progressbar):
        """
        :param progressbar: the progress bar
        :type progressbar: :class:`PlaybackProgressBar`
        """
        menu.ProviderMenu.__init__(self, 'progressbar-context-menu', progressbar)

        self._position = -1

    def get_context(self):
        """
        Retrieves the context
        """
        context = {'current-position': self._position}

        return context

    def popup(self, event):
        """
        Pops up the menu

        :param event: an event
        :type event: :class:`Gdk.Event`
        """
        self._position = event.x / self._parent.get_allocation().width

        menu.ProviderMenu.popup(self, event)


class MarkerContextMenu(menu.ProviderMenu):
    """
    Marker specific context menu
    """

    def __init__(self, markerbar):
        """
        :param markerbar: the marker capable progress bar
        :type markerbar: :class:`SeekProgressBar`
        """
        menu.ProviderMenu.__init__(self, 'playback-marker-context-menu', markerbar)

        self._markers = ()
        self._position = -1

    def regenerate_menu(self):
        """
        Builds the menu, with submenu if appropriate
        """
        for marker in self._markers:
            label = self._parent.get_label(marker)

            if label is None:
                continue

            markup_data = Pango.parse_markup(label, -1, '0')
            label_item = Gtk.MenuItem.new_with_mnemonic(markup_data[2])
            self.append(label_item)

            if len(self._markers) > 1:
                item_menu = Gtk.Menu()
                label_item.set_submenu(item_menu)
            else:
                item_menu = self
                label_item.set_sensitive(False)
                self.append(Gtk.SeparatorMenuItem())

            context = {
                'current-marker': marker,
                'selected-markers': self._markers,
                'current-position': self._position,
            }

            for item in self._items:
                i = item.factory(self, self._parent, context)
                item_menu.append(i)

        self.show_all()

    def popup(self, event, markers):
        """
        Pops up the menu

        :param event: an event
        :type event: :class:`Gdk.Event`
        :param markers: (m1, m2, ...)
        :type markers: (:class:`Marker`, ...)
        """
        self._markers = markers
        self._position = event.x / self._parent.get_allocation().width

        menu.ProviderMenu.popup(self, event)


class MoveMarkerMenuItem(menu.MenuItem):
    """
    Menu item allowing for movement of markers
    """

    def __init__(self, name, after, display_name=_('Move'), icon_name=None):
        menu.MenuItem.__init__(self, name, None, after)

        self._parent = None
        self._display_name = display_name
        self._icon_name = icon_name
        self._marker = None
        self._reset_position = -1

    def factory(self, menu, parent, context):
        """
        Generates the menu item
        """
        self._parent = parent

        item = Gtk.ImageMenuItem.new_with_mnemonic(self._display_name)

        if self._icon_name is not None:
            item.set_image(
                Gtk.Image.new_from_icon_name(self._icon_name, Gtk.IconSize.MENU)
            )

        item.connect('activate', self.on_activate, parent, context)

        parent.connect('button-press-event', self.on_parent_button_press_event)
        parent.connect('motion-notify-event', self.on_parent_motion_notify_event)
        parent.connect('focus-out-event', self.on_parent_focus_out_event)

        return item

    def move_begin(self, marker):
        """
        Captures the current marker for movement

        :param marker: the marker
        :type marker: :class:`Marker`
        :returns: whether a marker could be captured
        :rtype: bool
        """
        self.move_cancel()

        if marker is not None:
            self._marker = marker
            self._marker.props.state = Gtk.StateType.ACTIVE
            self._reset_position = marker.props.position
            self._parent.props.window.set_cursor(
                Gdk.Cursor.new(Gdk.CursorType.SB_H_DOUBLE_ARROW)
            )

            return True

        return False

    def move_update(self, position):
        """
        Moves the marker

        :param position: the current marker position
        :type position: float
        :returns: whether a marker could be moved
        :rtype: bool
        """
        if self._marker is not None:
            self._marker.props.position = position
            label = self._parent.get_label(self._marker)
            self._parent.set_tooltip_markup(label)

            return True

        return False

    def move_finish(self):
        """
        Finishes movement and releases the marker

        :returns: whether the movement could be finished
        :rtype: bool
        """
        if self._marker is not None:
            self._marker.props.state = Gtk.StateType.NORMAL
            self._marker = None
            self._reset_position = -1
            self._parent.props.window.set_cursor(None)

            return True

        return False

    def move_cancel(self):
        """
        Cancels movement and releases the marker

        :returns: whether the movement could be cancelled
        :rtype: bool
        """
        if self._marker is not None:
            self._marker.props.position = self._reset_position
            self._marker.props.state = Gtk.StateType.NORMAL
            self._marker = None
            self._reset_position = -1
            self._parent.props.window.set_cursor(None)

            return True

        return False

    def on_activate(self, widget, parent, context):
        """
        Starts movement of markers
        """
        self.move_begin(context.get('current-marker', None))

    def on_parent_button_press_event(self, widget, event):
        """
        Finishes or cancels movement of markers
        """
        if event.button == Gdk.BUTTON_PRIMARY:
            return self.move_finish()
        elif event.triggers_context_menu():
            return self.move_cancel()

        return False

    def on_parent_motion_notify_event(self, widget, event):
        """
        Moves markers
        """
        position = event.x / widget.get_allocation().width

        return self.move_update(position)

    def on_parent_focus_out_event(self, widget, event):
        """
        Cancels movement of markers
        """
        self.move_cancel()


class NewMarkerMenuItem(MoveMarkerMenuItem):
    """
    Menu item allowing for insertion
    and instant movement of a marker
    """

    def __init__(self, name, after):
        MoveMarkerMenuItem.__init__(self, name, after, _('New Marker'), 'list-add')

    def move_cancel(self):
        """
        Cancels movement and insertion of the marker

        :param parent: the parent
        :type parent: :class:`SeekProgressBar`
        :returns: whether the movement could be cancelled
        :rtype: bool
        """
        if self._marker is not None:
            remove_marker(self._marker)
            self._marker = None
            self._reset_position = -1
            self._parent.props.window.set_cursor(None)

            return True

        return False

    def on_activate(self, widget, parent, context):
        """
        Inserts a new marker and starts movement
        """
        context['current-marker'] = add_marker(context['current-position'])
        MoveMarkerMenuItem.on_activate(self, widget, parent, context)


# XXX: Example implementation only
# Bookmarks: "Add bookmark" (1 new marker)
# A-B-Repeat: "Repeat" (2 new marker, NW, NE)


def __create_progressbar_context_menu():
    items = []

    items.append(NewMarkerMenuItem('new-marker', []))

    for item in items:
        providers.register('progressbar-context-menu', item)


__create_progressbar_context_menu()


def __create_marker_context_menu():
    items = []

    def on_jumpto_item_activate(widget, name, parent, context):
        # parent.seek(context['current-marker'].props.position)
        position = context['current-marker'].props.position
        player.PLAYER.set_progress(position)

    def on_remove_item_activate(widget, name, parent, context):
        providers.unregister('playback-markers', context['current-marker'])

    items.append(
        menu.simple_menu_item(
            'jumpto-marker', [], _("_Jump to"), 'go-jump', on_jumpto_item_activate
        )
    )
    items.append(MoveMarkerMenuItem('move-marker', [items[-1].name]))
    items.append(
        menu.simple_menu_item(
            'remove-marker',
            [items[-1].name],
            _("_Remove Marker"),
            'list-remove',
            on_remove_item_activate,
        )
    )

    for item in items:
        providers.register('playback-marker-context-menu', item)


__create_marker_context_menu()


@GtkTemplate('ui', 'widgets', 'volume_control.ui')
class VolumeControl(Gtk.Box):
    """
    Encapsulates a button and a slider to
    control the volume indicating the current
    status via icon and tooltip
    """

    __gtype_name__ = 'VolumeControl'

    button, slider, button_image, slider_adjustment = GtkTemplate.Child.widgets(4)

    def __init__(self, player):
        Gtk.Box.__init__(self)
        self.init_template()

        self.button.add_events(Gdk.EventMask.KEY_PRESS_MASK | Gdk.EventMask.SCROLL_MASK)

        self.__volume_setting = '%s/volume' % player._name
        self.restore_volume = settings.get_option(self.__volume_setting, 1)
        self.icon_names = ['low', 'medium', 'high']
        self.__update(self.restore_volume)

        event.add_ui_callback(self.on_option_set, '%s_option_set' % player._name)

    def __update(self, volume):
        """
        Sets the volume level indicator
        """
        icon_name = 'audio-volume-muted'
        tooltip = _('Muted')

        if volume > 0:
            i = clamp(int(round(volume * 2)), 0, len(self.icon_names) - 1)
            icon_name = 'audio-volume-%s' % self.icon_names[i]
            # TRANSLATORS: Volume percentage
            tooltip = _('%d%%') % (volume * 100)
        else:
            volume = 0

        if volume == 1.0:
            tooltip = _('Full Volume')

        if volume > 0:
            self.button.set_active(False)

        self.button_image.set_from_icon_name(icon_name, Gtk.IconSize.BUTTON)
        self.button.set_tooltip_text(tooltip)
        self.slider.set_value(volume)
        self.slider.set_tooltip_text(tooltip)

    @GtkTemplate.Callback
    def on_scroll_event(self, widget, event):
        """
        Changes the volume on scrolling
        """
        page_increment = self.slider_adjustment.props.page_increment
        step_increment = self.slider_adjustment.props.step_increment
        value = self.slider.get_value()

        if event.direction in (Gdk.ScrollDirection.DOWN, Gdk.ScrollDirection.LEFT):
            if event.get_state() & Gdk.ModifierType.SHIFT_MASK:
                self.slider.set_value(value - page_increment)
            else:
                self.slider.set_value(value - step_increment)
            return True
        elif event.direction in (Gdk.ScrollDirection.UP, Gdk.ScrollDirection.RIGHT):
            if event.get_state() & Gdk.ModifierType.SHIFT_MASK:
                self.slider.set_value(value + page_increment)
            else:
                self.slider.set_value(value + step_increment)
            return True
        elif event.direction == Gdk.ScrollDirection.SMOOTH:
            delta = event.delta_x - event.delta_y
            if event.get_state() & Gdk.ModifierType.SHIFT_MASK:
                self.slider.set_value(value + delta * page_increment)
            else:
                self.slider.set_value(value + delta * step_increment)
            return True

        return False

    @GtkTemplate.Callback
    def on_button_toggled(self, button):
        """
        Mutes or unmutes the volume
        """
        if button.get_active():
            self.restore_volume = settings.get_option(self.__volume_setting, 1)
            volume = 0
        else:
            volume = self.restore_volume

        if self.restore_volume > 0:
            settings.set_option(self.__volume_setting, volume)

    @GtkTemplate.Callback
    def on_slider_value_changed(self, slider):
        """
        Stores the preferred volume
        """
        settings.set_option(self.__volume_setting, slider.get_value())

    @GtkTemplate.Callback
    def on_slider_key_press_event(self, slider, event):
        """
        Changes the volume on key press
        while the slider is focused
        """
        page_increment = slider.get_adjustment().props.page_increment
        step_increment = slider.get_adjustment().props.step_increment
        value = slider.get_value()

        if event.keyval == Gdk.KEY_Down:
            slider.set_value(value - step_increment)
            return True
        elif event.keyval == Gdk.KEY_Page_Down:
            slider.set_value(value - page_increment)
            return True
        elif event.keyval == Gdk.KEY_Up:
            slider.set_value(value + step_increment)
            return True
        elif event.keyval == Gdk.KEY_Page_Up:
            slider.set_value(value + page_increment)
            return True

        return False

    def on_option_set(self, event, sender, option):
        """
        Updates the volume indication
        """
        if option == self.__volume_setting:
            self.__update(settings.get_option(option, 1))


def playpause(player):
    if player.get_state() in ('playing', 'paused'):
        player.toggle_pause()
    else:
        from xlgui import main

        page = main.get_selected_playlist()
        if page:
            pl = page.playlist
            if len(pl) == 0:
                return
            try:
                idx = page.view.get_selected_paths()[0][0]
            except IndexError:
                idx = 0
            player.queue.set_current_playlist(pl)
            pl.current_position = idx
            player.queue.play(track=pl.current)


def PlayPauseMenuItem(name, player, after):
    def factory(name, after, player):
        if player.is_playing():
            icon_name = 'media-playback-pause'
            label = _("_Pause")
        else:
            icon_name = 'media-playback-start'
            label = _("P_lay")
        return menu.simple_menu_item(
            name, after, label, icon_name, callback=lambda *args: playpause(player)
        )

    return factory(name, after, player)


def NextMenuItem(name, player, after):
    return menu.simple_menu_item(
        name,
        after,
        _("_Next Track"),
        'media-skip-forward',
        callback=lambda *args: player.queue.next(),
    )


def PrevMenuItem(name, player, after):
    return menu.simple_menu_item(
        name,
        after,
        _("_Previous Track"),
        'media-skip-backward',
        callback=lambda *args: player.queue.prev(),
    )


def StopMenuItem(name, player, after):
    return menu.simple_menu_item(
        name,
        after,
        _("_Stop"),
        'media-playback-stop',
        callback=lambda *args: player.stop(),
    )



# Copyright (C) 2012  Mathias Brodala <info@noctus.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from gi import require_version

require_version('Atk', '1.0')

from gi.repository import Atk
from gi.repository import Gtk
from gi.repository import Gdk

from xl.nls import gettext as _
from xl import event, providers
from xlgui import main
from xlgui.widgets import notebook

EXAILE = None


def enable(exaile):
    """
    Enables the plugin
    """
    global EXAILE
    EXAILE = exaile
    if exaile.loading:
        event.add_callback(on_gui_loaded, 'gui_loaded')
    else:
        on_gui_loaded()


def disable(exaile):
    """
    Disables the plugin
    """
    global EXAILE
    EXAILE = None
    providers.unregister('main-panel-actions', MainMenuButton)


def on_gui_loaded(*args):
    """
    Creates the main menu button
    which takes care of the rest
    """
    if EXAILE.gui.panel_notebook.get_n_pages() == 0:
        raise Exception(_("This plugin needs at least one visible panel"))

    providers.register('main-panel-actions', MainMenuButton)


class MainMenuButton(Gtk.ToggleButton, notebook.NotebookAction):
    __gsignals__ = {}

    name = 'main-menu'
    position = Gtk.PackType.START

    def __init__(self, panel_notebook):
        """
        Adds the button to the main window
        and moves the main menu items
        """
        Gtk.ToggleButton.__init__(self)
        notebook.NotebookAction.__init__(self, panel_notebook)

        self.set_image(Gtk.Image.new_from_icon_name('exaile', Gtk.IconSize.BUTTON))
        self.set_tooltip_text(_('Main Menu'))
        self.set_focus_on_click(True)
        self.set_relief(Gtk.ReliefStyle.NONE)

        accessible = self.get_accessible()
        accessible.set_role(Atk.Role.MENU)
        accessible.set_name(_('Main Menu'))

        builder = main.mainwindow().builder

        # Move menu items of the main menu to the internal menu
        self.mainmenu = mainmenu = builder.get_object('mainmenu')
        self.menu = menu = Gtk.Menu()
        menu.attach_to_widget(self, None)
        menu.connect('deactivate', self.on_menu_deactivate)

        for menuitem in mainmenu:
            mainmenu.remove(menuitem)
            menu.add(menuitem)

        menu.show_all()
        self.show_all()

        self.connect('toggled', self.on_toggled)

        self.notebook_page_removed_connection = panel_notebook.connect(
            'page-removed', self.on_notebook_page_removed
        )

    def destroy(self):
        """
        Moves the main menu items back and
        removes the button from the main window
        """
        self.notebook.disconnect(self.notebook_page_removed_connection)

        menu = self.menu
        mainmenu = self.mainmenu
        for menuitem in menu:
            menu.remove(menuitem)
            mainmenu.add(menuitem)

        self.unparent()
        Gtk.Button.destroy(self)

    def get_menu_position(self, *_):
        """
        Positions the menu at the right of the button
        """
        # Origin includes window position and decorations
        _, x, y = self.props.window.get_origin()

        allocation = self.get_allocation()

        return (x + allocation.x + allocation.width, y + allocation.y, False)

    def do_button_press_event(self, e):
        """
        Pops out the menu upon click
        """
        if e.button == Gdk.BUTTON_PRIMARY:
            self.set_active(not self.get_active())

        return True

    def do_popup_menu(self):
        """
        Pops out the menu upon pressing
        the Menu or Shift+F10 keys
        """
        self.set_active(True)

        return True

    def on_toggled(self, button):
        """
        Pops out the menu upon button toggle
        """
        self.menu.popup(
            None, None, self.get_menu_position, None, 0, Gtk.get_current_event_time()
        )

    def on_menu_deactivate(self, menu):
        """
        Removes button activation upon menu popdown
        """
        self.set_active(False)

    def on_notebook_page_removed(self, notebook, *_):
        if notebook.get_n_pages() == 0:
            EXAILE.plugins.disable_plugin(__name__)



# Copyright (C) 2006 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 1, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import os
from xlgui.preferences import widgets
from xl.nls import gettext as _

name = _('Streamripper')
basedir = os.path.dirname(os.path.realpath(__file__))
ui = os.path.join(basedir, 'streamripper.ui')


class SavePreference(widgets.DirPreference):
    default = os.getenv('HOME')
    name = 'plugin/streamripper/save_location'


class PortPreference(widgets.Preference):
    default = '8888'
    name = 'plugin/streamripper/relay_port'


class FilePreference(widgets.CheckPreference):
    default = False
    name = 'plugin/streamripper/single_file'


class DeletePreference(widgets.CheckPreference):
    default = True
    name = 'plugin/streamripper/delete_incomplete'



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

"""
D-Bus interface for playback control, data query and others

Access through the ``/org/exaile/Exaile`` object which
implements the ``org.exaile.Exaile`` interface
"""

from collections import namedtuple
import logging
import sys

import dbus
import dbus.service
from gi.repository import Gio

# Be VERY careful what you import here! This module gets loaded even if
# we are just issuing a dbus command to a running instance, so we need
# to keep imports as light as possible.
from xl import event
from xl.nls import gettext as _

logger = logging.getLogger(__name__)


def check_dbus(bus, interface):
    obj = bus.get_object('org.freedesktop.DBus', '/org/freedesktop/DBus')
    dbus_iface = dbus.Interface(obj, 'org.freedesktop.DBus')
    avail = dbus_iface.ListNames()
    return interface in avail


def check_exit(options, args):
    """
    Check to see if dbus is running, and if it is, call the appropriate
    methods
    """
    iface = None
    if not options.NewInstance:
        # TODO: handle dbus stuff
        bus = dbus.SessionBus()
        if check_dbus(bus, 'org.exaile.Exaile'):
            remote_object = bus.get_object('org.exaile.Exaile', '/org/exaile/Exaile')
            iface = dbus.Interface(remote_object, 'org.exaile.Exaile')
            iface.TestService('testing dbus service')

            # Assume that args are files to be added to the current playlist.
            # This enables:    exaile PATH/*.mp3
            if args:
                # if '-' is the first argument then we look for a newline
                # separated list of filenames from stdin.
                # This enables:    find PATH -name *.mp3 | exaile -
                if args[0] == '-':
                    args = sys.stdin.read().split('\n')
                args = [Gio.File.new_for_commandline_arg(arg).get_uri() for arg in args]
                iface.Enqueue(args)

    if not iface:
        for command in [
            'GetArtist',
            'GetTitle',
            'GetAlbum',
            'GetLength',
            'GetRating',
            'SetRating',
            'IncreaseVolume',
            'DecreaseVolume',
            'Play',
            'Pause',
            'Stop',
            'Next',
            'Prev',
            'PlayPause',
            'StopAfterCurrent',
            'GuiToggleVisible',
            'CurrentPosition',
            'CurrentProgress',
            'GetVolume',
            'Query',
            'FormatQuery',
        ]:
            if getattr(options, command):
                return "command"
        return "continue"

    run_commands(options, iface)
    return "exit"


def run_commands(options, iface):
    """
    Actually invoke any commands passed in.
    """
    comm = False
    info_commands = {
        'GetArtist': 'artist',
        'GetTitle': 'title',
        'GetAlbum': 'album',
        'GetLength': '__length',
    }

    for command, attr in info_commands.items():
        if getattr(options, command):
            value = iface.GetTrackAttr(attr)
            if value is None:
                print(_('Not playing.'))
            else:
                print(value)
            comm = True

    argument_commands = (
        'IncreaseVolume',
        'DecreaseVolume',
        'SetRating',
        'Add',
        'ExportPlaylist',
    )

    for command in argument_commands:
        argument = getattr(options, command)
        if argument is not None:
            if command in ('IncreaseVolume', 'DecreaseVolume'):
                iface.ChangeVolume(
                    argument if command == 'IncreaseVolume' else -argument
                )
            else:
                print(getattr(iface, command)(argument))

            comm = True

    # Special handling for FormatQuery & FormatQueryTags
    format = options.FormatQuery
    if format is not None:
        print(
            iface.FormatQuery(
                format, options.FormatQueryTags or 'title,artist,album,__length'
            )
        )
        comm = True

    run_commands = (
        'Play',
        'Pause',
        'Stop',
        'Next',
        'Prev',
        'PlayPause',
        'StopAfterCurrent',
        'GuiToggleVisible',
        'ToggleMute',
    )

    for command in run_commands:
        if getattr(options, command):
            getattr(iface, command)()
            comm = True

    query_commands = (
        'CurrentPosition',
        'CurrentProgress',
        'GetRating',
        'GetVolume',
        'Query',
    )

    for command in query_commands:
        if getattr(options, command):
            print(getattr(iface, command)(argument))
            comm = True

    to_implement = ('GuiQuery',)
    for command in to_implement:
        if getattr(options, command):
            logger.warning("FIXME: command not implemented")
            comm = True

    if not comm:
        iface.GuiToggleVisible()


PlaybackStatus = namedtuple('PlaybackStatus', 'state progress position current')


class DbusManager(dbus.service.Object):
    """
    The dbus interface object for Exaile
    """

    def __init__(self, exaile):
        """
        Initializes the interface
        """
        self.exaile = exaile
        self.bus = dbus.SessionBus()
        self.bus_name = dbus.service.BusName('org.exaile.Exaile', bus=self.bus)
        dbus.service.Object.__init__(self, self.bus_name, '/org/exaile/Exaile')
        self.cached_track = ""
        self.cached_state = ""
        self.cached_volume = -1

    def _connect_signals(self):
        # connect events
        from xl import player

        event.add_callback(
            self.emit_state_changed, 'playback_player_end', player.PLAYER
        )
        event.add_callback(
            self.emit_state_changed, 'playback_track_start', player.PLAYER
        )
        event.add_callback(
            self.emit_state_changed, 'playback_toggle_pause', player.PLAYER
        )
        event.add_callback(self.emit_track_changed, 'tags_parsed', player.PLAYER)
        event.add_callback(self.emit_state_changed, 'playback_buffering', player.PLAYER)
        event.add_callback(self.emit_state_changed, 'playback_error', player.PLAYER)

    @dbus.service.method('org.exaile.Exaile', 's')
    def TestService(self, arg):
        """
        Just test the dbus object

        :param arg: anything
        """
        logger.debug(arg)

    @dbus.service.method('org.exaile.Exaile', None, 'b')
    def IsPlaying(self):
        """
        Determines if Exaile is playing / paused or not

        :returns: whether Exaile is playing / paused
        :rtype: boolean
        """
        from xl import player

        return bool(not player.PLAYER.is_stopped())

    @dbus.service.method('org.exaile.Exaile', 's', 's')
    def GetTrackAttr(self, attr):
        """
        Returns the value of a track tag

        :param attr: a track tag
        :type attr: string
        :returns: the tag value
        :rtype: string
        """
        from xl import player

        try:
            value = player.PLAYER.current.get_tag_raw(attr) or ''
        except (ValueError, TypeError, AttributeError):
            value = ''

        if isinstance(value, list):
            return "\n".join(value)
        return str(value)

    @dbus.service.method('org.exaile.Exaile', 'sv')
    def SetTrackAttr(self, attr, value):
        """
        Sets the value of a track tag

        :param attr: a track tag
        :type attr: string
        :param value: the tag value
        :type value: any
        """
        from xl import player

        try:
            player.PLAYER.current.set_tag_raw(attr, value)
        except (AttributeError, TypeError):
            pass

    @dbus.service.method('org.exaile.Exaile', None, 'i')
    def GetRating(self):
        """
        Returns the current track's rating

        :returns: the rating
        :rtype: int
        """
        try:
            rating = int(float(self.GetTrackAttr('__rating')))
        except ValueError:
            rating = 0

        return rating

    @dbus.service.method('org.exaile.Exaile', 'i')
    def SetRating(self, value):
        """
        Sets the current track's rating

        :param value: the new rating
        :type value: int
        """
        self.SetTrackAttr('__rating', value)
        event.log_event('rating_changed', self, value)

    @dbus.service.method('org.exaile.Exaile', 'i')
    def ChangeVolume(self, value):
        """
        Changes volume by the specified amount (in percent, can be negative)

        :param value: the new volume
        :type value: int
        """
        from xl import player

        player.PLAYER.set_volume(player.PLAYER.get_volume() + value)
        self.cached_volume = -1

    @dbus.service.method('org.exaile.Exaile')
    def ToggleMute(self):
        """
        Mutes or unmutes the volume
        """
        from xl import player

        if self.cached_volume == -1:
            self.cached_volume = player.PLAYER.get_volume()
            player.PLAYER.set_volume(0)
        else:
            player.PLAYER.set_volume(self.cached_volume)
            self.cached_volume = -1

    @dbus.service.method('org.exaile.Exaile', 'd')
    def Seek(self, value):
        """
        Seeks to the given position in seconds

        :param value: the position in seconds
        :type value: int
        """
        from xl import player

        player.PLAYER.seek(value)

    @dbus.service.method('org.exaile.Exaile')
    def Prev(self):
        """
        Jumps to the previous track
        """
        from xl import player

        player.QUEUE.prev()

    @dbus.service.method('org.exaile.Exaile')
    def Stop(self):
        """
        Stops playback
        """
        from xl import player

        player.PLAYER.stop()

    @dbus.service.method('org.exaile.Exaile')
    def Next(self):
        """
        Jumps to the next track
        """
        from xl import player

        player.QUEUE.next()

    @dbus.service.method('org.exaile.Exaile')
    def Play(self):
        """
        Starts playback
        """
        from xl import player

        player.QUEUE.play()

    @dbus.service.method('org.exaile.Exaile')
    def Pause(self):
        """
        Starts playback
        """
        from xl import player

        player.PLAYER.pause()

    @dbus.service.method('org.exaile.Exaile')
    def PlayPause(self):
        """
        Toggle Play or Pause
        """
        from xl import player

        if player.PLAYER.is_stopped():
            player.PLAYER.play(player.QUEUE.get_current())
        else:
            player.PLAYER.toggle_pause()

    @dbus.service.method('org.exaile.Exaile')
    def StopAfterCurrent(self):
        """
        Toggle stopping after current track
        """
        from xl import player

        player.QUEUE.stop_track = player.QUEUE.get_current()

    @dbus.service.method('org.exaile.Exaile', None, 's')
    def CurrentProgress(self):
        """
        Returns the progress into the current track (in percent)

        :returns: the current progress
        :rtype: string
        """
        from xl import player

        progress = player.PLAYER.get_progress()
        if progress == -1:
            return ""
        return str(int(progress * 100))

    @dbus.service.method('org.exaile.Exaile', None, 's')
    def CurrentPosition(self):
        """
        Returns the position inside the current track (as time)

        :returns: the current position
        :rtype: string
        """
        from xl import player

        progress = player.PLAYER.get_time()
        return '%d:%02d' % (progress // 60, progress % 60)

    @dbus.service.method('org.exaile.Exaile', None, 's')
    def GetVolume(self):
        """
        Returns the current volume level (in percent)

        :returns: the current volume
        :rtype: string
        """
        from xl import player

        return str(player.PLAYER.get_volume())

    def __get_playback_status(self, tags=["title", "artist", "album", "__length"]):
        """
        Retrieves the playback status

        :param tags: tags to retrieve from the current track
        :type tags: list
        :returns: the playback status
        :rtype: :class:`xl.xldbus.PlaybackStatus`
        """
        status = PlaybackStatus(state='stopped', current=None, progress=0, position=0)

        from xl import player

        current_track = player.QUEUE.get_current()
        if current_track is not None and not player.PLAYER.is_stopped():
            current = {}
            # Make tags unique
            tags = list(set(tags))

            for tag in tags:
                current[tag] = self.GetTrackAttr(tag)

            # Special handling for internal tags
            if "__length" in tags:
                from xl.formatter import LengthTagFormatter

                current["__length"] = LengthTagFormatter.format_value(
                    self.GetTrackAttr('__length')
                )

            status = PlaybackStatus(
                state=player.PLAYER.get_state(),
                progress=self.CurrentProgress(),
                position=self.CurrentPosition(),
                current=current,
            )

        return status

    @dbus.service.method('org.exaile.Exaile', None, 's')
    def Query(self):
        """
        Returns information about the currently playing track

        :returns: information about the current track
        :rtype: string
        """
        status = self.__get_playback_status()

        if status.current is None or status.state == "stopped":
            return _('Not playing.')

        result = _(
            'status: %(status)s, title: %(title)s, artist: %(artist)s,'
            ' album: %(album)s, length: %(length)s,'
            ' position: %(progress)s%% [%(position)s]'
        ) % {
            'status': status.state,
            'title': status.current["title"],
            'artist': status.current["artist"],
            'album': status.current["album"],
            'length': status.current["__length"],
            'progress': status.progress,
            'position': status.position,
        }

        return result

    @dbus.service.method('org.exaile.Exaile', 'ss', 's')
    def FormatQuery(self, format, tags):
        """
        Returns the current playback state including
        information about the currently playing track

        :param format: the desired output format (currently only "json")
        :type format: string
        :param tags: tags to retrieve from the current track
        :type tags: string
        :returns: the formatted information or an empty
                  string if requesting an unknown format
        :rtype: string
        """
        # TODO: Elaborate the use of providers here
        if format == 'json':
            import json

            status = self.__get_playback_status(tags.split(","))

            return json.dumps(status._asdict())

        return ''

    @dbus.service.method('org.exaile.Exaile', None, 's')
    def GetVersion(self):
        """
        Returns the version of Exaile

        :returns: the application version
        :rtype: string
        """
        return self.exaile.get_version()

    @dbus.service.method('org.exaile.Exaile', 's')
    def PlayFile(self, filename):
        """
        Plays the specified file

        :param filename: the path to start playing
        :type filename: string
        """
        self.exaile.gui.open_uri(filename)

    @dbus.service.method('org.exaile.Exaile', 'as')
    def Enqueue(self, locations):
        """
        Adds the tracks at the specified locations
        to the current playlist

        :param locations: locations to enqueue
        :type locations: iterable
        """
        from xl import player
        from xlgui import get_controller

        controller = get_controller()

        play = False

        # Allows for playing first item
        if player.PLAYER.is_stopped():
            play = True

        for location in locations:
            controller.open_uri(location, play=play)
            play = False

    @dbus.service.method('org.exaile.Exaile', 's')
    def Add(self, location):
        """
        Adds the tracks at the specified
        location to the collection

        :param location: where to add tracks from
        :type location: string
        """
        from xl import trax

        tracks = trax.get_tracks_from_uri(location)
        self.exaile.collection.add_tracks(tracks)

    @dbus.service.method('org.exaile.Exaile', 's')
    def ExportPlaylist(self, location):
        """
        Exports the current playlist
        to the specified location

        :param location: where to save the playlist at
        :type location: string
        """
        from xl import player, playlist

        if player.QUEUE.current_playlist is not None:
            playlist.export_playlist(player.QUEUE.current_playlist, location)

    @dbus.service.method('org.exaile.Exaile')
    def GuiToggleVisible(self):
        """
        Toggles visibility of the GUI, if possible
        """
        self.exaile.gui.main.toggle_visible(bringtofront=True)

    @dbus.service.method('org.exaile.Exaile', None, 'ay')
    def GetCoverData(self):
        """
        Returns the data of the cover image of the playing track, or
        an empty string if there is no cover available.

        :returns: the cover data
        :rtype: binary data
        """
        from xl import covers, player

        cover = covers.MANAGER.get_cover(player.PLAYER.current, set_only=True)
        if not cover:
            cover = ''
        return cover

    @dbus.service.method('org.exaile.Exaile', None, 's')
    def GetState(self):
        """
        Returns the surrent verbatim state (unlocalized)

        :returns: the player state
        :rtype: string
        """
        from xl import player

        return player.PLAYER.get_state()

    @dbus.service.signal('org.exaile.Exaile')
    def StateChanged(self):
        """
        Emitted when state change occurs: 'playing' 'paused' 'stopped'
        """
        pass

    @dbus.service.signal('org.exaile.Exaile')
    def TrackChanged(self):
        """
        Emitted when track change occurs.
        """
        pass

    def emit_state_changed(self, type, player, object):
        """
        Called from main to emit signal
        """
        new_state = player.get_state()
        if self.cached_state != new_state:
            self.cached_state = new_state
            self.StateChanged()

    def emit_track_changed(self, type, player, object):
        """
        Called from main to emit signal
        """
        new_track = self.GetTrackAttr('__loc')
        if self.cached_track != new_track:
            self.cached_track = new_track
            self.TrackChanged()


# vim: et sts=4 sw=4



#!/usr/bin/env python3

"""Launcher for Exaile on Windows"""

import builtins
import msvcrt
import sys
import os
from ctypes import windll

# Provide dummy sys.stdout and sys.stderr, which are None in noconsole/windowed
# mode (launched using pythonw.exe, or PyInstaller build).
if sys.stdout is None:
    sys.stdout = open(os.devnull, "w")
if sys.stderr is None:
    sys.stderr = open(os.devnull, "w")


__builtin__open = __builtins__.open


def __open_inheritance_hack(*args, **kwargs):
    """Make file handles not inheritable, that way we can restart on the fly"""
    # From http://www.virtualroadside.com/blog/index.php/2013/02/06/problems-with-file-descriptors-being-inherited-by-default-in-python/
    result = __builtin__open(*args, **kwargs)
    handle = msvcrt.get_osfhandle(result.fileno())
    windll.kernel32.SetHandleInformation(handle, 1, 0)
    return result


builtins.open = __open_inheritance_hack


def error(message1, message2=None, die=True):
    import logging

    """Show error message and exit.

    If two message arguments are supplied, the first one will be used as title.
    If `die` is true, exit after showing the message.
    """
    logging.error(message1 + (('\r\n\r\n' + message2) if message2 else ''))
    if sys.stdout.isatty():
        if die:
            print("\r\n[Press Enter to exit.]", file=sys.stderr)
            input()
    else:
        import ctypes

        if not message2:
            message1, message2 = message2, message1
        ctypes.windll.user32.MessageBoxW(None, message2, message1, 0x10)
    if die:
        sys.exit(1)


def main():
    import logging

    logging.basicConfig(
        level=logging.INFO, datefmt="%H:%M:%S", format="%(levelname)-8s: %(message)s"
    )

    aio_message = (
        "\r\n\r\nPlease run the 'All-In-One PyGI/PyGObject for "
        "Windows Installer' and ensure that the following are selected:"
        "\r\n\r\n"
        "* GTK+ 3.x\r\n"
        "* GStreamer 1.x and the gst-plugins package(s)\r\n"
        "\r\n"
        "The 'All-In-One PyGI/PyGObject for Windows Installer' "
        "can be downloaded at\r\n"
        "https://sourceforge.net/projects/pygobjectwin32/\r\n\r\n"
        "See README.Windows for more information."
    )

    try:
        import gi
    except Exception:
        error("PyGObject not found", "PyGObject could not be imported. " + aio_message)
    else:
        logging.info("PyGObject works")
    try:
        gi.require_version('Gtk', '3.0')
        from gi.repository import Gtk
    except Exception:
        error("GTK+ not found", "GTK+ could not be imported. " + aio_message)
    else:
        logging.info("GTK+ works")

    try:
        gi.require_version('Gst', '1.0')
        from gi.repository import Gst
    except Exception:
        error("GStreamer not found", "GStreamer could not be imported. " + aio_message)
    else:
        logging.info("GStreamer works")

    try:
        import mutagen
    except Exception:
        error(
            "Mutagen not found",
            "The Python module Mutagen could not be imported. For download "
            "and installation instructions see "
            "https://mutagen.readthedocs.io/en/latest/\r\n\r\n"
            "See README.Windows for more information.",
        )
    else:
        logging.info("Mutagen works")

    # disable the logging before starting exaile.. otherwise it gets
    # configured twice and we get double the log messages!
    logging = None
    del sys.modules['logging']

    try:
        sys.argv[1:1] = ['--startgui', '--no-dbus', '--no-hal']
        import exaile

        exaile.main()
    except Exception:
        import traceback

        error("Error while running exaile.main()", traceback.format_exc())


if __name__ == '__main__':
    main()

# vi: et sts=4 sw=4 ts=4



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

from xl import covers, settings, xdg
from xl.nls import gettext as _
from xlgui.preferences import widgets

name = _('Covers')
icon = 'image-x-generic'
ui = xdg.get_data_path('ui', 'preferences', 'cover.ui')


class TagCoverFetching(widgets.CheckPreference):
    default = True
    name = 'covers/use_tags'


class LocalCoverFetching(widgets.CheckPreference):
    default = True
    name = 'covers/use_localfile'


class LocalFilePreferredNamesPreference(widgets.Preference, widgets.CheckConditional):
    default = ['front', 'cover', 'album']
    name = 'covers/localfile/preferred_names'
    condition_preference_name = 'covers/use_localfile'

    def __init__(self, preferences, widget):
        widgets.Preference.__init__(self, preferences, widget)
        widgets.CheckConditional.__init__(self)

    def _get_value(self):
        """
        Converts the string value to a list
        """
        return [v.strip() for v in widgets.Preference._get_value(self).split(',')]

    def _set_value(self):
        """
        Converts the list to a string value
        """
        self.widget.set_text(', '.join(settings.get_option(self.name, self.default)))


class CoverOrderPreference(widgets.OrderListPreference):
    """
    This little preference item shows kind of a complicated preference
    widget in action.  The defaults are dynamic.
    """

    name = 'covers/preferred_order'

    def __init__(self, preferences, widget):
        self.default = covers.MANAGER._get_methods()
        widgets.OrderListPreference.__init__(self, preferences, widget)

    def _set_value(self):
        self.model.clear()
        for item in self.default:
            self.model.append([item.name])

    def apply(self):
        if widgets.OrderListPreference.apply(self):
            covers.MANAGER.set_preferred_order(self._get_value())
        return True


class AutomaticCoverFetching(widgets.CheckPreference):
    default = True
    name = 'covers/automatic_fetching'



'''
Simple program that uses the 'bpmdetect' GStreamer plugin to detect
the BPM of a song, and outputs that to console.

Requires GStreamer 1.x, PyGObject 1.x, and gst-plugins-bad

Copyright (C) 2015 Dustin Spicuzza

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2 as
published by the Free Software Foundation.
'''

import inspect
import sys
from gi.repository import Gst, GObject, Gio


def autodetect_supported():
    return Gst.ElementFactory.make('bpmdetect', None) is not None


def detect_bpm(uri, on_complete):
    """
    Detects the BPM of a song using GStreamer's bpmdetect plugin

    .. note:: The plugin emits the BPM at various times during
              song processing, but the bpm detector accumulates
              the results so this will only return the last
              result.
    """

    bpm = [None]

    def _on_message(bus, msg):
        if msg.type == Gst.MessageType.TAG:
            tags = msg.parse_tag()

            # Discard tags already set on the file
            if tags.n_tags() > 1:
                return

            v = tags.get_value_index('beats-per-minute', 0)
            try:
                v = float(v)
            except Exception:
                return

            if v > 0:
                bpm[0] = v

        elif msg.type == Gst.MessageType.ERROR:
            playbin.set_state(Gst.State.NULL)

            gerror, debug_info = msg.parse_error()
            if gerror:
                on_complete(None, gerror.message.rstrip("."))
            else:
                on_complete(None, debug_info)

        elif msg.type == Gst.MessageType.EOS:
            playbin.set_state(Gst.State.NULL)
            on_complete(bpm[0], None)

    audio_sink = Gst.Bin.new('audiosink')

    # bpmdetect doesn't work properly with more than one channel,
    # see https://bugzilla.gnome.org/show_bug.cgi?id=751457
    cf = Gst.ElementFactory.make('capsfilter', None)
    cf.props.caps = Gst.Caps.from_string('audio/x-raw,channels=1')

    fakesink = Gst.ElementFactory.make('fakesink', None)
    fakesink.props.sync = False
    fakesink.props.signal_handoffs = False

    bpmdetect = Gst.ElementFactory.make('bpmdetect', None)
    if bpmdetect is None:
        on_complete(None, "GStreamer BPM detection plugin not found")
        return

    audio_sink.add(cf)
    audio_sink.add(bpmdetect)
    audio_sink.add(fakesink)

    cf.link(bpmdetect)
    bpmdetect.link(fakesink)

    audio_sink.add_pad(Gst.GhostPad.new('sink', cf.get_static_pad('sink')))

    playbin = Gst.ElementFactory.make('playbin', None)
    playbin.props.audio_sink = audio_sink
    playbin.props.flags &= ~1  # Unset GST_PLAY_FLAG_VIDEO

    bus = playbin.get_bus()
    bus.add_signal_watch()
    bus.connect('message', _on_message)

    playbin.props.uri = uri
    playbin.set_state(Gst.State.PLAYING)


if __name__ == '__main__':
    Gst.init(None)

    if len(sys.argv) != 2:
        print("Usage: %s filename\n" % sys.argv[0])
        print(inspect.cleandoc(__doc__))
        exit(1)

    uri = Gio.File.new_for_commandline_arg(sys.argv[1]).get_uri()
    print("Processing file: ", uri)

    def on_complete(bpm, error):
        if error is not None:
            print("ERROR:", error)
        else:
            print("BPM:", bpm)

        ml.quit()

    detect_bpm(uri, on_complete)

    ml = GObject.MainLoop()
    ml.run()



#!/usr/bin/env python3

import re
import os
from os.path import abspath, dirname, join, exists
import sys

v_re1 = re.compile(r"^Version='(.*)'$")
v_re2 = re.compile(r'^Version="(.*)"$')


def _find_bad_plugin_versions(plugins_dir):
    for pname in os.listdir(plugins_dir):
        p = join(plugins_dir, pname, 'PLUGININFO')
        if exists(p):
            with open(p) as fp:
                contents = fp.read()

            s = contents.splitlines()
            for i, l in enumerate(s):
                m = v_re1.match(l) or v_re2.match(l)
                if m:
                    vv = m.group(1)
                    if vv != v:
                        yield pname, p, vv, s[:i] + s[i + 1:]

                    break
            else:
                print("Warning: no version found in %s" % pname)


def check(v, plugins):
    retval = 0
    for pname, _, vv, _ in plugins:
        print("ERROR: Plugin %s version is %s, not %s" % (pname, vv, v))
        retval = 1

    return retval


def fixversion(v, plugins):
    v = "Version='%s'" % v
    for _, p, _, contents in plugins:
        contents = '\n'.join([v] + contents) + '\n'
        with open(p, 'w') as fp:
            fp.write(contents)


if __name__ == '__main__':
    exaile_dir = abspath(join(dirname(__file__), '..'))
    plugins_dir = join(exaile_dir, 'plugins')

    os.environ['EXAILE_DIR'] = exaile_dir
    sys.path.insert(0, exaile_dir)

    import xl.version
    v = '%s' % (xl.version.__version__.split("-")[0].split("+")[0])

    arg = sys.argv[1] if len(sys.argv) > 1 else ''

    if arg == 'check':
        retval = check(v, _find_bad_plugin_versions(plugins_dir))
    elif arg == 'fix':
        retval = fixversion(v, _find_bad_plugin_versions(plugins_dir))
    else:
        print('Usage: %s check|fix' % sys.argv[0])
        retval = 1

    exit(retval)



# Copyright (C) 2009-2010 Aren Olson
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

from gi.repository import Gtk
from gi.repository import GLib

from xl.nls import gettext as _
from xl import common, event, player, providers
from xl import settings as xl_settings
from xl import lyrics as xl_lyrics
from xlgui import guiutil, panel
from xlgui.preferences import lyrics as lyricsprefs


class LyricsPanel(panel.Panel):
    # public variable for xlgui.panel.Panel
    ui_info = ('lyrics.ui', 'LyricsPanel')

    def __init__(self, parent, name):
        panel.Panel.__init__(self, parent, name, _('Lyrics'))

        self.__lyrics_found = []
        self.__css_provider = Gtk.CssProvider()

        WIDGET_LIST = [
            'lyrics_top_box',
            'refresh_button',
            'refresh_button_stack',
            'refresh_icon',
            'refresh_spinner',
            'track_text',
            'scrolled_window',
            'lyrics_text',
            'lyrics_source_label',
            'track_text_buffer',
            'lyrics_text_buffer',
        ]
        for name in WIDGET_LIST:
            setattr(
                self,
                '_' + LyricsPanel.__name__ + '__' + name,
                self.builder.get_object(name),
            )
        self.__initialize_widgets()

        event.add_ui_callback(self.__on_playback_track_start, 'playback_track_start')
        event.add_ui_callback(self.__on_track_tags_changed, 'track_tags_changed')
        event.add_ui_callback(self.__on_playback_player_end, 'playback_player_end')
        event.add_ui_callback(
            self.__on_lyrics_search_method_added, 'lyrics_search_method_added'
        )
        event.add_ui_callback(self.__on_option_set, 'plugin_lyricsviewer_option_set')

        self.__update_lyrics()

    def __initialize_widgets(self):
        lyrics_methods_combo = LyricsMethodsComboBox()
        self.__lyrics_top_box.pack_start(lyrics_methods_combo, True, True, 0)
        lyrics_methods_combo.connect('changed', self.__on_combo_active_changed)
        lyrics_methods_combo.show()
        self.__lyrics_methods_combo = lyrics_methods_combo

        track_text_style = Gtk.CssProvider()
        style_context = self.__track_text.get_style_context()
        style_context.add_provider(
            track_text_style, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
        )
        track_text_style.load_from_data(b"textview {font-weight: bold; }")

        style_context = self.__lyrics_text.get_style_context()
        style_context.add_provider(
            self.__css_provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
        )

        lyricsprefs.DEFAULT_FONT = (
            self.__lyrics_text.get_default_attributes().font.to_string()
        )
        # trigger initial setup through options
        self.__on_option_set(None, xl_settings, 'plugin/lyricsviewer/lyrics_font')

        self.__refresh_button.connect('clicked', self.__on_refresh_button_clicked)

    def __on_option_set(self, _event, settings, option):
        if option == 'plugin/lyricsviewer/lyrics_font':
            pango_font_str = settings.get_option(option)
            css_from_pango = guiutil.css_from_pango_font_description(pango_font_str)
            data_str = "textview { " + css_from_pango + "; }\n"
            self.__css_provider.load_from_data(data_str.encode('utf-8'))

    def __on_lyrics_search_method_added(self, _eventtype, _lyrics, _provider):
        self.__update_lyrics()

    def __on_track_tags_changed(self, _eventtype, track, tags):
        if player.PLAYER.current == track and tags & {"artist", "title"}:
            self.__update_lyrics()

    def __on_playback_track_start(self, _eventtype, _player, _data):
        self.__update_lyrics()

    def __on_playback_player_end(self, _eventtype, _player, _data):
        self.__update_lyrics()

    def __on_refresh_button_clicked(self, _button):
        """
        Called when the refresh button is clicked
        """
        self.__update_lyrics(refresh=True)

    def __on_combo_active_changed(self, _combobox):
        """
        Catches when the user selects an item of the combo.
        """
        if self.__lyrics_found:
            self.__update_lyrics_text()

    def __update_lyrics(self, refresh=False):
        self.__track_text_buffer.set_text("")
        self.__lyrics_text_buffer.set_text("")
        self.__lyrics_source_label.set_text("")
        self.__lyrics_found = []
        if player.PLAYER.current:
            self.__set_top_box_widgets(False)
            self.__get_lyrics(player.PLAYER.current, refresh)
        else:
            self.__lyrics_text_buffer.set_text(_('Not playing.'))
            self.__set_top_box_widgets(False, True)

    @common.threaded
    def __get_lyrics(self, track, refresh=False):
        lyrics_found = []
        track_text = ''
        try:
            try:
                track_text = (
                    track.get_tag_raw('artist')[0]
                    + " - "
                    + track.get_tag_raw('title')[0]
                )
            except Exception:
                raise xl_lyrics.LyricsNotFoundException
            lyrics_found = xl_lyrics.MANAGER.find_all_lyrics(track, refresh)
        except xl_lyrics.LyricsNotFoundException:
            lyrics_found = []
        finally:
            self.__get_lyrics_finish(track, track_text, lyrics_found)

    @common.idle_add()
    def __get_lyrics_finish(self, track, track_text, lyrics_found):
        '''Only called from __get_lyrics thread, thunk to ui thread'''

        if track != player.PLAYER.current:
            return

        self.__lyrics_found = lyrics_found

        self.__track_text_buffer.set_text(track_text)
        self.__update_lyrics_text()
        self.__set_top_box_widgets(True)

    def __update_lyrics_text(self):
        lyrics = _("No lyrics found.")
        source = ""
        url = ""
        if self.__lyrics_found:
            index, selected_method = self.__lyrics_methods_combo.get_active_item()
            for name, i_lyrics, i_source, i_url in self.__lyrics_found:
                if name == selected_method or index == 0:
                    lyrics, source, url = i_lyrics, i_source, i_url
                    break
        self.__lyrics_text_buffer.set_text(lyrics)

        if url != "":
            url_text = '<a href="' + url + '">' + source + '</a>'
            self.__lyrics_source_label.set_markup(_("Source: ") + url_text)
        else:
            self.__lyrics_source_label.set_text("")

    def __set_top_box_widgets(self, state, init=False):
        if state or init:
            self.__refresh_spinner.stop()
            self.__refresh_button_stack.set_visible_child(self.__refresh_icon)
        else:
            self.__refresh_button_stack.set_visible_child(self.__refresh_spinner)
            self.__refresh_spinner.start()

        self.__refresh_button.set_sensitive(state)
        self.__lyrics_methods_combo.set_sensitive(state)


class LyricsMethodsComboBox(Gtk.ComboBoxText, providers.ProviderHandler):
    """
    An extended Gtk.ComboBox class.
    Shows lyrics methods search registered
    """

    def __init__(self):
        Gtk.ComboBoxText.__init__(self)
        providers.ProviderHandler.__init__(self, 'lyrics')

        self.model = self.get_model()
        # Default value, any registered lyrics provider
        self.append_text(_("Any source"))

        for provider in self.get_providers():
            self.append_text(provider.display_name)

        self.set_active(0)

    def remove_item(self, name):
        index = self.search_item(name)
        if index:
            GLib.idle_add(self.remove, index)
            return True
        return False

    def append_item(self, name):
        if not self.search_item(name):
            GLib.idle_add(self.append_text, name)
            return True
        return False

    def search_item(self, name):
        index = 0
        for item in self.model:
            if item[0] == name:
                return index
            index += 1
        return False

    def get_active_item(self):
        return self.get_active(), self.get_active_text()

    def on_provider_added(self, provider):
        self.append_item(provider.display_name)

    def on_provider_removed(self, provider):
        self.remove_item(provider.display_name)
